/*
    Copyright (C) 2001-2002 by theKompany.com <www.thekompany.com>
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    In addition, as a special exception, theKompany.com gives permission
    to link the code of this program with the tkwidgets library (or with
    modified versions of tkwidgets that use the same license as tkwidgets),
    and distribute linked combinations including the two.  You must obey
    the GNU General Public License in all respects for all of the code used
    other than tkwidgets.  If you modify this file, you may extend this
    exception to your version of the file, but you are not obligated to do so.
    If you do not wish to do so, delete this exception statement from your
    version.

    This license grants you the ability to use tkwidgets with Aethera only
    and may not be used outside of Aethera.
    See also http://www.thekompany.com/products/license.txt for details.
*/

#include "smtpclient.h"
#include "hostname.h"

#include <appstatus.h>
#include <addressclass.h>
#include <servernotifier.h>
#include <indexclass.h>
#include <qregexp.h>
#include <messagefactory.h>
#include <messagedevice.h>
#include <mimecodec.h>
#include <orb.h>

// SSL support
#include <sslsocket.h>
#include <ssldevice.h>

using namespace tkcSSL;

class SMTPClient::Private
{
  public:
  
    Private(SMTPClient*, Account*);
    virtual ~Private();
    
    SSLSocket * socket_;

    QStringList queue_;

    Account       * account_;
    
    IndexClass    * currentIndex_;
    MessageDevice * messageDevice_;
    
    AppStatus     * status;
    unsigned        messages;
    
    SMTPClient::State state_;
    Account::SendServerType server;
    Account::AuthType auth;
    
    QTimer          idleTimer_;
    QStringList     addressList_;
    
  private:
    
    SMTPClient * parent_;
};

SMTPClient::Private::Private(SMTPClient * parent, Account * account)
  : socket_         (0L),
    account_        (account),
    currentIndex_   (0L),
    messageDevice_  (0L),
    status          (0L),
    messages        (0),
    state_          (SMTPClient::Idle),
    server          (Account::Sendmail),
    auth            (Account::NONE),
    parent_         (parent)
{
  socket_ = new SSLSocket();
  
  if( account_->s_ssl )
  {
    SSLDevice * sslDevice = new SSLDevice(QSocketDevice::Stream, SSLDevice::SSL23);
    socket_->setSocketDevice(sslDevice, true);
    if( !sslDevice->initSSL() )
    {
      qDebug("SMTPClient : Error initSSL()");
    }
  }
  else if( account_->s_tls )
  {
    // TODO: STARTTLS
  }
  
  status = AppStatus::instance();
}

SMTPClient::Private::~Private()
{
  if( socket_ )
  {
    delete socket_;
    socket_ = 0L;
  }
}

SMTPClient::SMTPClient(Account * account, QObject * parent, const char * name)
  : QObject(parent, name)
{
  d = new Private(this, account);
  
  connect
    (
      d->socket_,
      SIGNAL(hostFound()),
      SLOT(slotHostFound())
    );

  connect
    (
      d->socket_,
      SIGNAL(connected()),
      SLOT(slotConnected())
    );

  connect
    (
      d->socket_,
      SIGNAL(connectionClosed()),
      SLOT(slotConnectionClosed())
    );

  connect
    (
      d->socket_,
      SIGNAL(readyRead()),
      SLOT(slotReadyRead())
    );

  connect
    (
      d->socket_,
      SIGNAL(bytesWritten(int)),
      SLOT(slotBytesWritten(int))
    );

  connect
    (
      d->socket_,
      SIGNAL(error(int)),
      SLOT(slotError(int))
    );

  connect
    (
      &d->idleTimer_,
      SIGNAL(timeout()),
      SLOT(slotIdle())
    );
}

SMTPClient::~SMTPClient()
{
  while (!d->queue_.isEmpty())
  {
    QString indexURL = d->queue_.first();
    
    IndexClass * index = ObjectRequestBroker::thisInstance()->indexReference( indexURL );
    if( index )
      index->unlock();
    
    d->queue_.pop_front();
  }
  
  if(d)
  {
    delete d;
    d = 0L;
  }
}

  void
SMTPClient::startSend(IndexClass * index)
{
  if( index )
  {
    index->lock();
    d->queue_ += index->url();
    startNextSend();
  }
}

  void
SMTPClient::startSend(const QStringList & messages)
{
  d->messages = 0;
  IndexClass * index = 0L;
  ObjectRequestBroker * orb = ObjectRequestBroker::thisInstance();
  
  for(QStringList::ConstIterator it=messages.begin(); it!=messages.end(); ++it)
  {
    index = orb->indexReference( (*it) );
    if( index && !index->isLocked() )
    {
      index->lock();
      d->queue_ += (*it);
    }
  }
  
  if( !d->queue_.isEmpty() )
  {
    d->messages = d->queue_.count();
    d->status->message(AppStatus::Server, AppStatus::Progress, AppStatus::Waiting,
      "Sending...", 0, d->messages);
    
#ifdef SMTP_DEBUG
    qDebug("Scheduling %u messages to send", d->queue_.count());
#endif

    startNextSend();
  }
}

  void
SMTPClient::startNextSend()
{
  if (Idle != d->state_)
  {
    qDebug("SMTPClient::startNextSend: Not idle");
    return;
  }

  if (d->queue_.isEmpty())
  {
    qDebug("SMTPClient::startNextSend: Queue empty");
    emit(finished(this));
    return;
  }

  d->status->message(AppStatus::Server, AppStatus::Progress, AppStatus::Waiting,
    "Sending...", d->messages - d->queue_.count(), d->messages);
    
#ifdef SMTP_DEBUG
  qDebug("SMTPClient::startNextSend: Queue has %d items", d->queue_.count());
#endif

  QString indexURL = d->queue_.first();
  d->currentIndex_ = ObjectRequestBroker::thisInstance()->indexReference( indexURL );
  d->queue_.pop_front();

  if( !d->currentIndex_ )
  {
    qDebug("current sending mail has no index");
    emit(finished(this));
    return;
  }

  // A message device seems to be a proxy object for a message.
  MessageDevice * messageDevice = new MessageDevice(d->currentIndex_);
// MessageFactory::thisInstance()->getMessage(d->currentIndex_);

  if (0 == messageDevice)
  {
    qWarning("SMTPClient::startNextSend: MessageDevice is 0");
    return;
  }

  d->messageDevice_ = messageDevice;

  // A message descriptor seems to be message meta-info (address lists, etc.)
  messageDevice->loadDescriptor(false /* not fast - whatever that means */);

  MessageDescriptor & descriptor = messageDevice->getDescriptor();

  d->addressList_ = descriptor.to;

  d->addressList_ += descriptor.cc;
  d->addressList_ += descriptor.bcc;

  ServerNotifier::thisInstance()
    ->ioProgress(QString("smtp:sending:1:%2").arg(d->queue_.count()));

  setState(WaitingForConnection);

#ifdef SMTP_DEBUG
  qDebug("SMTPClient::startNextSend: Host: %s", d->account_->smtp.latin1());
  qDebug("SMTPClient::startNextSend: Port: %d", d->account_->s_port);
#endif

  d->socket_->connectToHost(d->account_->smtp, d->account_->s_port);

  // TODO: Less coarse-grained timeout - see hints in RFC (4.5.3.2).

  resetIdleTimer();
}

  void
SMTPClient::nextSend()
{
//  if (Idle == d->state_)
//  {
//    qDebug("SMTPClient::nextSend: Idle");
//    sendComplete();
//    return;
//  }

  if (d->queue_.isEmpty())
  {
    qDebug("SMTPClient::nextSend: Queue empty");
    sendComplete();
    return;
  }

  QString indexURL = d->queue_.first();
  d->currentIndex_ = ObjectRequestBroker::thisInstance()->indexReference( indexURL );
  d->queue_.pop_front();

  if( !d->currentIndex_ )
  {
    qDebug("current sending mail has no index");
    emit(finished(this));
    return;
  }

//  qDebug("SMTPClient::startNextSend: id: %s", d->currentIndex_->getID().latin1());

  // A message device seems to be a proxy object for a message.
  MessageDevice * messageDevice = new MessageDevice(d->currentIndex_);
// MessageFactory::thisInstance()->getMessage(d->currentIndex_);

  if (0 == messageDevice)
  {
    qWarning("SMTPClient::startNextSend: MessageDevice is 0");
    return;
  }

  d->messageDevice_ = messageDevice;

  // A message descriptor seems to be message meta-info (address lists, etc.)
  messageDevice->loadDescriptor(false /* not fast - whatever that means */);

  MessageDescriptor & descriptor = messageDevice->getDescriptor();

  d->addressList_ = descriptor.to;

  d->addressList_ += descriptor.cc;
  d->addressList_ += descriptor.bcc;

  ServerNotifier::thisInstance()
    ->ioProgress(QString("smtp:sending:1:%2").arg(d->queue_.count()));

  sendMailFrom();

  resetIdleTimer();
}

  void
SMTPClient::sendComplete()
{
  d->status->message(AppStatus::Server, AppStatus::Text, AppStatus::Good, "Finished");
  
  if (0 != d->currentIndex_)
  {
    // Reset everything.
    // TODO: Refactor, putting these in a proper State class.
    d->currentIndex_->unlock();
    d->currentIndex_ = 0;
  }
  d->messageDevice_ = 0;
  d->addressList_.clear();
  d->socket_->close();
  setState(Idle);
  d->idleTimer_.stop();
  emit(finished(this));
}

  void
SMTPClient::slotHostFound()
{
#ifdef SMTP_DEBUG
  qDebug("SMTPClient::slotHostFound");
#endif

  d->status->message(AppStatus::Server, AppStatus::Text, AppStatus::Waiting, "Host found...");
  
  resetIdleTimer();
}

  void
SMTPClient::slotConnected()
{
#ifdef SMTP_DEBUG
  qDebug("SMTPClient::slotConnected");
#endif

  d->status->message(AppStatus::Server, AppStatus::Text, AppStatus::Waiting, "Connected...");
  
  resetIdleTimer();
}

  void
SMTPClient::slotConnectionClosed()
{
#ifdef SMTP_DEBUG
  qDebug("SMTPClient::slotConnectionClosed");
#endif

  d->status->message(AppStatus::Server, AppStatus::Text, AppStatus::Warning, "Connection closed...");

  if (Idle != d->state_)
  {
    qDebug("Wasn't idle !");
  }

  sendComplete();
}

  void
SMTPClient::slotReadyRead()
{
#ifdef SMTP_DEBUG
  qDebug("SMTPClient::slotReadyRead");
#endif

  resetIdleTimer();

  while (d->socket_->canReadLine())
  {
    QString line = d->socket_->readLine();

    if (line.length() < 3)
    {
      qDebug("Bogus server response: `%s'", line.latin1());
    }
    else
    {
      uint code = line.left(3).toUInt();

      switch (d->state_)
      {
        case WaitingForConnection:

          if (220 != code)
          {
            qDebug("WaitingForConnection but response is %d", code);
            sendComplete();
            return;
          }

//          qDebug("WaitingForConnection: ok (%d)", code);
          if( Account::NONE == d->account_->s_auth )
            sendHelo();
          else
            sendEhlo();

          break;

        case WaitingForHeloResponse:
          if (code < 200 || code > 299)
          {
            qDebug("WaitingForHeloResponse but response is %d", code);
            sendComplete();
            return;
          }
          handlerHelo(line);
          break;

        case WaitingForAuthResponse:
          handlerAuth(line);
          break;
          
        case WaitingForAuthChallenge:
          handlerAuthChallenge(line);
          break;
          
        case WaitingForMailFromResponse:

          if (code < 200 || code > 299)
          {
            qDebug("WaitingForMailFromResponse but response is %d", code);
            sendComplete();
            return;
          }

//          qDebug("WaitingForMailFromResponse: ok (%d)", code);
          sendRcptTo();

          break;

        case WaitingForRcptToResponse:

          if (code < 200 || code > 299)
          {
            // Policy: We're only sending via a relay, so it should accept
            // all addresses we give it. If any are rejected, drop out.

            qDebug("WaitingForRcptToResponse but response is %d", code);
            sendComplete();
            return;
          }

//          qDebug("WaitingForRcptToResponse: ok (%d)", code);

          // More addresses to tell the server about ?

          if (!d->addressList_.isEmpty())
          {
            // Yes, so tell the server about the next one.
            sendRcptTo();
          }
          else
          {
            // No, so we tell the server we want to send the message now.
            sendDataBegin();
          }

          break;

        case WaitingForDataBeginResponse:

          if (354 != code)
          {
            qDebug("WaitingForDataBeginResponse but response is %d", code);
            sendComplete();
            return;
          }

//          qDebug("WaitingForDataBeginResponse : ok (%d)", code);
          sendData();

          break;

        case WaitingForDataEndResponse:

          if (250 != code)
          {
            qDebug("WaitingForDataEndResponse but response is %d", code);
            sendComplete();
            return;
          }

//          qDebug("WaitingForDataEndResponse : ok (%d)", code);
          
          d->currentIndex_->unlock();
          emit( messageDelivered(d->currentIndex_) );
          
          // d->currentIndex_ will disappear here.
          d->currentIndex_ = 0;
          nextSend();
          return;

          break;

        default:

          qDebug("Don't know what to do with response: `%s'", line.latin1());
      }
    }
  }
}

  void
SMTPClient::slotBytesWritten(int)
{
//  qDebug("SMTPClient::slotBytesWritten(%d)", byteCount);

  resetIdleTimer();
}

  void
SMTPClient::slotError(int err)
{
  qDebug("SMTPClient::slotError(%d)", err);

  if (Idle != d->state_)
  {
    qDebug("Wasn't idle !");
  }

  sendComplete();
}

  void

SMTPClient::slotIdle()
{
#ifdef SMTP_DEBUG
  qDebug("SMTPClient::slotIdle");
#endif

  if (Idle == d->state_)
  {
    qDebug("Was already idle");
    return;
  }

  sendComplete();
}

  void
SMTPClient::sendHelo()
{
  writeLine("HELO " + System::hostname());
  setState(WaitingForHeloResponse);
}

  void
SMTPClient::sendEhlo()
{
  writeLine("EHLO " + System::hostname());
  setState(WaitingForHeloResponse);
}

  void
SMTPClient::sendMailFrom()
{
  Q_ASSERT(0 != d->messageDevice_);

	MessageDescriptor & descriptor = d->messageDevice_->getDescriptor();

  writeLine("MAIL FROM: " + AddressClass(descriptor.from.latin1()).email);

  setState(WaitingForMailFromResponse);
}

  void
SMTPClient::sendRcptTo()
{
  Q_ASSERT(0 != d->messageDevice_);

  Q_ASSERT(!d->addressList_.isEmpty());

  QString address = d->addressList_.first();

  d->addressList_.pop_front();

  writeLine("RCPT TO: " + AddressClass(address.latin1()).email);
  setState(WaitingForRcptToResponse);
}

  void
SMTPClient::sendDataBegin()
{
  writeLine("DATA");
  setState(WaitingForDataBeginResponse);
}

  void
SMTPClient::sendData()
{
  // Write message data.
  QTextStream str(d->socket_);

  QString messageData = d->messageDevice_->rfc822Message();
  char lastChar = 0;
  char currentChar = 0;
  unsigned uLength = messageData.length();
  QString line;

  for(unsigned i=0; i<uLength; ++i)
  {
    currentChar = messageData[(int)i].latin1();
    
    switch( currentChar )
    {
      case '\n':
        lastChar = currentChar;
        line += "\r\n";
        str << line;
        line = QString::null;
        break;
      case '.':
        if( '\n' == lastChar )
        {
          lastChar = currentChar;
          line += "..";
          break;
        }
      default :
        lastChar = currentChar;
        line += currentChar;
        break;
    }
  }
  
  if( !line.isEmpty() )
    str << line;

  str << "\r\n.\r\n";
  
  setState(WaitingForDataEndResponse);
}

  void
SMTPClient::sendQuit()
{
  QTextStream str(d->socket_);
  writeLine("QUIT");
}

  Account *
SMTPClient::account()
{
  return d->account_;
}

  void
SMTPClient::resetIdleTimer()
{
  d->idleTimer_.start(60000 * 10 /* 10 minutes */, true);
}

  void
SMTPClient::writeLine(const QString & s)
{
//  qDebug("SMTPClient::writeLine(`%s')", s.latin1());
  QTextStream str(d->socket_);
  str << s << "\r\n";
}

  void
SMTPClient::setState(SMTPClient::State state)
{
//  qDebug
//    (
//      "SMTPClient: State change: %s -> %s",
//      stateToString(d->state_).latin1(),
//      stateToString(state).latin1()
//    );

  d->state_ = state;
}

  QString
SMTPClient::stateToString(SMTPClient::State state) const
{
  QString s = "Unknown";

  switch (state)
  {
    case Idle:
      s = "Idle";
      break;

    case WaitingForConnection:
      s = "WaitingForConnection";
      break;

    case WaitingForHeloResponse:
      s = "WaitingForHeloResponse";
      break;

    case WaitingForAuthResponse:
      s = "WaitingForAuthResponse";
      break;

    case WaitingForAuthChallenge:
      s = "WaitingForAuthChallenge";
      break;

    case WaitingForMailFromResponse:
      s = "WaitingForMailFromResponse";
      break;

    case WaitingForRcptToResponse:
      s = "WaitingForRcptToResponse";
      break;

    case WaitingForDataBeginResponse:
      s = "WaitingForDataBeginResponse";
      break;

    case WaitingForDataEndResponse:
      s = "WaitingForDataEndResponse";
      break;

    default:
      break;
  }

  return s;
}

void SMTPClient::handlerHelo(const QString & line)
{
  if ("250" != line.left(3))
  {
    qDebug( tr("Error while determining server capabilities. Server said: %1").arg(line) );
    sendComplete();
    return;
  }
  else if ('-' == line[3])
  {
    parseCapability(line.mid(4));
  }
  else if (' ' == line[3])
  {
    parseCapability(line.mid(4));

    if(d->account_->s_auth != Account::NONE)
    {
      // try to send it even if auth capability is empty
      if( d->auth == Account::NONE ||
          d->auth == d->account_->s_auth )
      {
        sendAuth();
      }
      else
      {
        qDebug("Wanted auth type is not supported");
        sendComplete();
      }
    }
    else
    {
      sendMailFrom();
    }
  }
}

void SMTPClient::sendAuth()
{
  QString authTypeString = authTypeEnumToString(d->auth);

#ifdef SMTP_DEBUG
  qDebug("Using auth type: %s", authTypeString.ascii());
#endif

  if(Account::PLAIN == d->auth)
  {
    QCString user = d->account_->s_user.utf8().data();
    QCString pass = d->account_->s_password.utf8().data();

    int userLength = user.length();
    int passLength = pass.length();

    QCString result;

    result.resize(2 * userLength + passLength + 2);

    {
      for (int i = 0; i <= userLength; i++)
        result[i + userLength + 1] = result[i] = user[i];
    }

    {
      for (int i = 0; i < passLength; i++)
        result[i +  2 * userLength + 2] = pass[i];
    }

    QCString encodedResult( MIMECodec::encodeBase64(result, false) );

    QString plainString(encodedResult);

    writeLine("AUTH PLAIN " + plainString);
    d->state_ = WaitingForAuthResponse;
  }
  else
  {
    writeLine("AUTH " + authTypeString);
    d->state_ = WaitingForAuthChallenge;
  }
}

void SMTPClient::handlerAuth(const QString & line)
{
  if ("235" != line.left(3))
  {
    qDebug("SMTPClient::handlerAuth error: %s", line.latin1());
    sendComplete();
    return;
  }

  sendMailFrom();
}

void SMTPClient::handlerAuthChallenge(const QString & line)
{
  if ("334" != line.left(3))
  {
    qDebug("SMTPClient::handlerAuthChallenge : %s", line.latin1());
    sendComplete();
    return;
  }

  QString data(line.mid(4).stripWhiteSpace());

  switch(d->auth)
  {
    case Account::LOGIN:
      handleLoginChallenge(data);
      break;
    case Account::PLAIN:
      handlePlainChallenge(data);
      break;
    case Account::CRAM_MD5:
      handleCramMD5Challenge(data);
      break;
    case Account::DIGEST_MD5:
      handleDigestMD5Challenge(data);
      break;
    case Account::NONE:
      qDebug("SMTPClient::handlerAuthChallenge : wrong auth");
      break;
  }
}

void SMTPClient::parseCapability(const QString & line)
{
  if("AUTH" == line.left(4).upper() && 
      d->account_->s_auth != Account::NONE )
  {
    QStringList authTypeTokenList = QStringList::split(' ', line);
    authTypeTokenList.remove(authTypeTokenList.begin());

    QStringList::ConstIterator it;

    for (it = authTypeTokenList.begin(); it != authTypeTokenList.end(); ++it)
    {
      QString authTypeString((*it).stripWhiteSpace());

#ifdef SMTP_DEBUG
      qDebug("Auth Supports: `%s'", authTypeString.ascii());
#endif

      Account::AuthType authType = authTypeStringToEnum(authTypeString);

      if(Account::NONE == authType)
      {
        qDebug("Don't know about auth type `%s'", authTypeString.ascii());
      }
      else if(d->account_->s_auth == authType)
      {
        d->auth = d->account_->s_auth;
      }
    }
  }
}

Account::AuthType SMTPClient::authTypeStringToEnum(const QString & _s)
{
  QString s(_s.lower());

  if ("login" == s)
    return Account::LOGIN;

  else if ("plain" == s)
    return Account::PLAIN;

  else if ("cram-md5" == s)
    return Account::CRAM_MD5;

  else if ("digest-md5" == s)
    return Account::DIGEST_MD5;

  return Account::NONE;
}

QString SMTPClient::authTypeEnumToString(Account::AuthType type)
{
  switch(type)
  {
    case Account::LOGIN:
      return "LOGIN";
      break;

    case Account::PLAIN:
      return "PLAIN";
      break;

    case Account::CRAM_MD5:
      return "CRAM-MD5";
      break;

    case Account::DIGEST_MD5:
      return "DIGEST-MD5";
      break;

    default:
      return "NONE";
      break;
  }
}

void SMTPClient::handleLoginChallenge(const QString & data)
{
  QCString s( MIMECodec::decodeBase64( QCString(data.latin1()) ) );

#ifdef SMTP_DEBUG
  qDebug("handleLoginChallenge() : `%s'", s.data());
#endif

  if( "username:" == s.lower() )
  {
    writeLine( MIMECodec::encodeBase64( QCString(d->account_->s_user.latin1()), false ) );
  }
  else if( "password:" == s.lower() )
  {
    d->state_ = WaitingForAuthResponse;
    
    writeLine( MIMECodec::encodeBase64( QCString(d->account_->s_password.latin1()), false ) );
  }
}

void SMTPClient::handlePlainChallenge(const QString &)
{
  qDebug("handlePlainChallenge: not implemented");
  sendComplete();
}

void SMTPClient::handleCramMD5Challenge(const QString &)
{
  qDebug("handleCramMD5Challenge: not implemented");
  sendComplete();
}

void SMTPClient::handleDigestMD5Challenge(const QString &)
{
  qDebug("handleDigestMD5Challenge: not implemented");
  sendComplete();
}

