/***************************************************************************
                 ssldevice.cpp  -  QSocketDevice with SSL support
                             -------------------
    begin                : Tue Apr 23 2002
    copyright            : (C) 2002 by theKompany.com
    author               : Eugen Constantinescu
    email                : eug@thekompany.com
 ***************************************************************************/

#include <cassert>

#include "ssldevice.h"

// #define SSL_DEBUG

#include <errno.h>

using namespace tkcSSL;

namespace tkcSSL
{
  class SSLDevicePrivate
  {
    public:
    
      SSLDevicePrivate(SSLDevice::ConnectionType conType = SSLDevice::NOSSL);
      ~SSLDevicePrivate();
      
      bool isSSL();
      bool isTLS();

#ifndef NO_OPENSSL
      SSL *_ssl;
      SSL_CTX *_ctx;
      SSL_METHOD *_method;
#endif
      
      SSLBase *sslBase;
      SSLDevice::ConnectionType _conType;
      
      SSLDevice::ConnectionState _state;
  };
};

// Private member
SSLDevicePrivate::SSLDevicePrivate(SSLDevice::ConnectionType conType)
  :
#ifndef NO_OPENSSL
    _ssl(0L),
    _ctx(0L),
    _method(0L),
#endif
    sslBase(0L),
    _conType(conType),
    _state( SSLDevice::CLOSE )
{
  sslBase = SSLBase::ref();
}

SSLDevicePrivate::~SSLDevicePrivate()
{
  _state = SSLDevice::CLOSE;
}

bool SSLDevicePrivate::isSSL()
{
#ifdef SSL_DEBUG
  qDebug("isSSL");
#endif

  if( _conType != SSLDevice::NOSSL )
    return true;
  
#ifdef SSL_DEBUG
  qDebug("isSSL - very bad");
#endif

  return false;
}

bool SSLDevicePrivate::isTLS()
{
  if( _conType == SSLDevice::TLS )
    return true;
  
  return false;
}

// SSL Device
SSLDevice::SSLDevice( Type type, SSLDevice::ConnectionType conType)
  : QSocketDevice(type)
{
  d = new SSLDevicePrivate(conType);
}

SSLDevice::SSLDevice( int socket, Type type, SSLDevice::ConnectionType conType)
  : QSocketDevice(socket, type)
{
  d = new SSLDevicePrivate(conType);
}

SSLDevice::~SSLDevice()
{
  close();
  
  if(d)
  {
    delete d;
    d = 0L;
  }
}
      
bool SSLDevice::open( int mode )
{
#ifdef SSL_DEBUG
  qDebug("SSLDevice::open");
#endif

  if( QSocketDevice::open(mode) )
  {
    if( d->isSSL() && d->sslBase->isOpenSSL() )
      return initSSL();
    
    return true;
  }

  return false;  
}

void SSLDevice::close()
{
  QSocketDevice::close();

  if( d->isSSL() && d->sslBase->isOpenSSL() )
    closeSSL();
}

bool SSLDevice::connect( const QHostAddress &addr, Q_UINT16 port)
{
  return QSocketDevice::connect(addr, port);
}


#if QT_VERSION < 300
int SSLDevice::bytesAvailable() const
#else
Q_LONG SSLDevice::bytesAvailable() const
#endif
{
#ifndef NO_OPENSSL
  if( d->_state != SSLDevice::CONNECTED )
    return -1;
  
  int iRet = d->sslBase->SSL_pending(d->_ssl);
  
  // check the internal buffer
  if( !iRet )
  {
    iRet = QSocketDevice::bytesAvailable();
  }
  
#ifdef SSL_DEBUG
  qDebug("SSLDevice::bytesAvailable = %d", iRet);
#endif
  
  return iRet;
#else
  return -1;
#endif
}

#if QT_VERSION < 300
int SSLDevice::readBlock( char *data, uint maxlen )
#else
Q_LONG SSLDevice::readBlock( char *data, Q_ULONG maxlen )
#endif
{

#ifdef SSL_DEBUG
  qDebug("SSLDevice::readBlock");
#endif

#ifndef NO_OPENSSL
  if( d->_state != SSLDevice::CONNECTED )
    return -1;
  return d->sslBase->SSL_read(d->_ssl, data, maxlen);
#else
  return -1;
#endif
}

#if QT_VERSION < 300
int SSLDevice::writeBlock( const char *data, uint len )
#else
Q_LONG SSLDevice::writeBlock( const char *data, Q_ULONG len )
#endif
{
#ifdef SSL_DEBUG
  qDebug("SSLDevice::writeBlock");
#endif

#ifndef NO_OPENSSL
  if( d->_state != SSLDevice::CONNECTED )
    return -1;
  return d->sslBase->SSL_write(d->_ssl, data, len);
#else
  return -1;
#endif
}

bool SSLDevice::initSSL()
{
#ifndef NO_OPENSSL
#ifdef SSL_DEBUG
  qDebug("initSSL");
#endif

  if( d->_state != SSLDevice::CLOSE )
    return false;

  switch(d->_conType)
  {
    case SSLDevice::SSL2 :
      d->_method = d->sslBase->SSLv2_client_method();
      break;
    case SSLDevice::SSL3 :
      d->_method = d->sslBase->SSLv3_client_method();
      break;
    case SSLDevice::SSL23 :
      d->_method = d->sslBase->SSLv23_client_method();
      break;
    case SSLDevice::TLS :
      d->_method = d->sslBase->TLSv1_client_method();
      break;
    default:
      assert(0);
      break;
  }
  if( NULL == d->_method )
    return false;
    
  d->_ctx = d->sslBase->SSL_CTX_new( d->_method );
  if( NULL == d->_ctx )
    return false;
  
  d->_state = SSLDevice::INIT;
  return true;
#else
  return false;
#endif
}

bool SSLDevice::connectSSL()
{
  if( d->isSSL() && d->sslBase->isOpenSSL() )
  {
#ifndef NO_OPENSSL
    qDebug("SSLDevice : ssl connect \n---------------\n ");
    if( d->_state != SSLDevice::INIT )
      return false;

    d->_ssl = d->sslBase->SSL_new( d->_ctx );
    if( !d->_ssl )
      return false;
        
#ifdef _OLD_EAY_
    if( d->isTLS() )
      SSL_CTX_set_options(d->_ctx, SSL_OP_ALL/*SSL_OP_NO_TLSv1*/);
    else
      SSL_CTX_set_options(d->_ctx, SSL_OP_ALL);
#else
    if( d->isTLS() )
      d->sslBase->SSL_set_options(d->_ssl, SSL_OP_NO_TLSv1);
    else
      d->sslBase->SSL_set_options(d->_ssl, SSL_OP_ALL);
#endif
      
    if( !(d->sslBase->SSL_set_fd(d->_ssl, socket())) )
      return false;

    if( !(d->sslBase->SSL_connect(d->_ssl)) )
      return false;
// TODO
//      else
//        getConnectionInfo();
#else
      return false;
#endif
  }
    
  d->_state = SSLDevice::CONNECTED;
    
  return true;  
}

void SSLDevice::closeSSL()
{
#ifndef NO_OPENSSL
  switch( d->_state )
  {
    case SSLDevice::CONNECTED :
      if( d->_ssl )
        d->sslBase->SSL_shutdown( d->_ssl );
    case SSLDevice::INIT :
      if( d->_ssl )
      {
        d->sslBase->SSL_free( d->_ssl );
        d->_ssl = 0L;
      }
      if( d->_ctx )
      {
        d->sslBase->SSL_CTX_free( d->_ctx );
        d->_ctx = 0L;
      }
    case SSLDevice::CLOSE :
    default:
      break;
  }
  d->_method = 0L;
  
  d->_state = SSLDevice::CLOSE;
#endif
}

